//
// Copyright (c) 2012, MindFusion LLC - Bulgaria.
//

using System;
using System.Collections;
using System.ComponentModel;
using System.Drawing;
using System.Globalization;
using System.Reflection;
using System.Windows.Forms;

using MindFusion.Diagramming;
using MindFusion.Diagramming.Layout;
using MindFusion.Diagramming.WinForms;
using MindFusion.Drawing;

using Brush = MindFusion.Drawing.Brush;


namespace MindFusion.Diagramming.WinForms.Samples.CS.ClassDiagram
{
	public partial class MainForm : Form
	{
		public MainForm()
		{
			InitializeComponent();

			_pureImages = new Image[]
				{
					_imageClass.Image,
					_imageDelegate.Image,
					_imageEnum.Image,
					_imageEnumField.Image,
					_imageEvent.Image,
					_imageField.Image,
					_imageInterface.Image,
					_imageMethod.Image,
					_imagePMethod.Image,
					_imagePProperty.Image,
					_imageProperty.Image,
					_imageStatic.Image,
					_imageStruct.Image,
					_imageNamespace.Image,
					_imagePField.Image
				};

			diagram.Selection.Style = SelectionStyle.Frame;
            diagram.BackBrush = new MindFusion.Drawing.SolidBrush(Color.FromArgb(224, 233, 233));
		}

		private void _menuLoad_Click(object sender, EventArgs e)
		{
			if (_openFileDialog.ShowDialog(this) == DialogResult.OK)
			{
				// Create the TypeName <-> TableNode map
				_workMap = new Hashtable();

				// Create tables nodes and fill the tree
				ProcessAssembly(_openFileDialog.FileName);

				// Generate links between classes
				GenerateLinks();

				// Arrange the diagram
				Arrange();
			}
		}

		private void _menuExit_Click(object sender, EventArgs e)
		{
			Close();
		}

		private void ProcessAssembly(string assemblyFile)
		{
			// Clear the tree
			_tree.Nodes.Clear();

			// Clear diagram
			diagram.ClearAll();

			// Load the assembly
			Assembly assembly;

			try
			{
				assembly = Assembly.LoadFrom(assemblyFile);
			}
			catch (Exception exception)
			{
				MessageBox.Show(this, "Failed to load assembly.\nReason: \"" +
					exception.Message + "\"", "Class Diagrammer",
					MessageBoxButtons.OK, MessageBoxIcon.Error);
				return;
			}

			if (assembly == null)
			{
				MessageBox.Show(this, "Failed to load assembly.", "ClassDiagrammer",
					MessageBoxButtons.OK, MessageBoxIcon.Error);
				return;
			}

			// Get exported types
			Type[] types = assembly.GetExportedTypes();

			// Sort the exported types
			ArrayList sortedTypes = new ArrayList(types);
			sortedTypes.Sort(new TypeComparer());

			// Populate the tree view with assembly information
			foreach (Type type in sortedTypes)
			{
				// Add the namespace of the type
				TreeNode node = AddNamespace(type.Namespace);

				// Add the type itself
				AddType(node, type);
			}

			// Update tables' bounds
			foreach (DiagramNode node in diagram.Nodes)
			{
				TableNode t = node as TableNode;
				if (t == null)
					continue;

				int rows = Math.Min(8, t.RowCount);
				float height = 7 + rows * 5;

				t.Bounds = new RectangleF(
					t.Bounds.X, t.Bounds.Y,
					t.Bounds.Width, height);
			}

			// Combine enums & delegates
			DiagramNodeCollection enums = new DiagramNodeCollection();
			DiagramNodeCollection delegates = new DiagramNodeCollection();
			foreach (DiagramNode node in diagram.Nodes)
			{
				TableNode t = node as TableNode;
				if (t == null)
					continue;

				int index = (t.Tag as TreeNode).ImageIndex;
				if (index == ImageEnum)
					enums.Add(t);
				else if (index == ImageDelegate)
					delegates.Add(t);
			}

			BuildGroup(enums, "Enums", ":enums");
			BuildGroup(delegates, "Delegates", ":delegates");
		}

		private void BuildGroup(DiagramNodeCollection tables, string name, object tag)
		{
			if (tables.Count > 0)
			{
				int side = (int)Math.Ceiling(Math.Sqrt(tables.Count));
				int rows = (int)Math.Ceiling((float)tables.Count / side);
				float singleWidth = tables[0].Bounds.Width;
				float width = 5 * (side + 1) + singleWidth * side;
				float height = 5 * (rows + 1) + 5;

				float[] rowHeights = new float[rows];
				for (int i = 0; i < tables.Count; i++)
					rowHeights[i / side] = Math.Max(tables[i].Bounds.Height, rowHeights[i / side]);

				foreach (float h in rowHeights)
					height += h;

				//ShapeNode b = new ShapeNode(diagram);
				//b.Bounds = new RectangleF(0, 0, width, height);
				//diagram.Nodes.Add(b);
				ContainerNode b = diagram.Factory.CreateContainerNode(0, 0, width, height);

				int index = 0;
				foreach (TableNode e in tables)
				{
					//e.AttachTo(b, AttachToNode.TopLeft);
					b.Add(e);

					float x = (index % side) * (5 + singleWidth);
					float y = 0;

					for (int r = 0; r < index / side; r++)
						y += 5 + rowHeights[r];

					e.Bounds = new RectangleF(x + 5, y + 10,
						e.Bounds.Width, e.Bounds.Height);

					index++;
				}

				b.Caption = name;
				//b.TextFormat = sf;
				b.HandlesStyle = HandlesStyle.HatchHandles3;
				b.ShadowOffsetX = 0;
				b.ShadowOffsetY = 0;
				b.Font = new Font("Verdana", 12, FontStyle.Bold);
				b.IgnoreLayout = true;
				b.Tag = tag;
				b.UpdateBounds();
			}
		}

		/// <summary>
		/// Walks the tree and creates the appropriate tables in the diagram.
		/// </summary>
		private void GenerateLinks()
		{
			// Create relation between types in the assembly
			foreach (DiagramNode node in diagram.Nodes)
			{
				TableNode t = node as TableNode;
				if (t == null)
					continue;

				TreeNode treeNode = t.Tag as TreeNode;
				Type type = treeNode.Tag as Type;
				Type baseType = type.BaseType;
				if (baseType == null)
					continue;

				string baseTypeName = baseType.Namespace + Type.Delimiter + baseType.Name;

				TableNode parent = _workMap[baseTypeName] as TableNode;
				if (parent != null)
					diagram.Links.Add(new DiagramLink(diagram, parent, t));
			}
		}

		/// <summary>
		/// Arranges the diagram.
		/// </summary>
		private void Arrange()
		{
			TreeLayout layout = new TreeLayout();

			layout.Type = TreeLayoutType.Centered;
			layout.LinkStyle = TreeLayoutLinkType.Cascading3;
			layout.LevelDistance = 10;
			layout.KeepGroupLayout = true;

			layout.Arrange(diagram);

			// Place enums and delegates at the end
			DiagramNode enums = diagram.FindNode(":enums");
			DiagramNode delegates = diagram.FindNode(":delegates");

			float x = 0;
			foreach (DiagramNode node in diagram.Nodes)
			{
				TableNode t = node as TableNode;
				if (t == null)
					continue;

				if (t.MasterGroup == null)
					x = Math.Max(x, t.Bounds.Right);
			}

			if (enums != null)
			{
				enums.Move(x + 5, 5);
				x = enums.Bounds.Right;
			}

			if (delegates != null)
			{
				delegates.Move(x + 5, 5);
			}

			diagram.ResizeToFitItems(5);
		}

		private TreeNode AddNamespace(string namespaceName)
		{
			string[] names = namespaceName.Split(new char[] { '.' });

			int index = ImageNamespace;
			TreeNodeCollection nodes = _tree.Nodes;
			TreeNode node = null;

			foreach (string name in names)
			{
				node = null;
				foreach (TreeNode n2 in nodes)
				{
					if (n2.Text == name)
					{
						node = n2;
						break;
					}
				}

				if (node == null)
				{
					// Find where to insert the new node
					int pos = nodes.Count;
					for (int i = 0; i < nodes.Count; i++)
					{
						TreeNode n3 = nodes[i];
						if (DefaultComparer.Compare(n3.Text, name) > 0)
							pos = i;
					}

					node = new TreeNode(name, index, index);
					nodes.Insert(pos, node);
				}

				nodes = node.Nodes;
			}

			return node;
		}

		private TreeNode AddType(TreeNode parent, Type type)
		{
			if (type.MemberType == MemberTypes.NestedType)
			{
				Stack stack = new Stack();

				Type tempType = type;
				while (tempType.MemberType == MemberTypes.NestedType)
				{
					tempType = tempType.DeclaringType;
					stack.Push(tempType);
				}

				TreeNode tempParent = parent;
				while (stack.Count > 0)
					tempParent = AddTypeNoContents(tempParent, stack.Pop() as Type);

				return AddTypeWithContents(tempParent, type);
			}
			else
			{
				return AddTypeWithContents(parent, type);
			}
		}

		/// <summary>
		/// Adds the specified type to the tree without
		/// adding the type contents.
		/// </summary>
		private TreeNode AddTypeNoContents(TreeNode parent, Type type)
		{
			// Do not add if already in
			foreach (TreeNode node in parent.Nodes)
			{
				if (node.Text == type.Name)
					return node;
			}

			// Create a table for the type
			TableNode t = new TableNode(diagram);
			t.Bounds = new RectangleF(10, 10, 40, 10);
			t.Caption = "<b>" + type.Name + "</b>";
			t.RedimTable(3, 0);
			t.ToolTip = type.Namespace + Type.Delimiter + type.Name;
			t.OffsetHeaderRows = true;
			diagram.Nodes.Add(t);

			int icon = -1;
			if (type.IsClass)
			{
				if (type.IsSubclassOf(typeof(Delegate)))
				{
					icon = ImageDelegate;

					t.Brush = DelegateBrush;
				}
				else
				{
					icon = ImageClass;

					// Set brush and tooltip
					t.Brush = ClassBrush;

					if (!type.BaseType.Equals(typeof(Object)))
					{
						t.ToolTip += "\nextends " + type.BaseType.Namespace +
							Type.Delimiter + type.BaseType.Name;
					}
				}
			}
			else if (type.IsInterface)
			{
				icon = ImageInterface;

				t.Brush = InterfaceBrush;
			}
			else if (type.IsEnum)
			{
				icon = ImageEnum;

				t.Brush = EnumBrush;
			}
			else if (type.IsValueType)
			{
				icon = ImageStruct;

				t.Brush = StructBrush;
			}

			TreeNode child = new TreeNode(type.Name, icon, icon);
			child.Tag = type;
			parent.Nodes.Add(child);

			t.Tag = child;
			t.CustomDraw = CustomDraw.Additional;
			t.AnchorPattern = AnchorPattern.TopInBottomOut;

			_workMap[type.Namespace + Type.Delimiter + type.Name] = t;

			return child;
		}

		/// <summary>
		/// Adds the specified type to the tree without recursing
		/// upwards if the type is nested.
		/// </summary>
		private TreeNode AddTypeWithContents(TreeNode parent, Type type)
		{
			// Add type item
			TreeNode child = AddTypeNoContents(parent, type);

			// Get the table that represents the type
			TableNode table = _workMap[type.Namespace + Type.Delimiter + type.Name] as TableNode;

			// Add contents
			switch (child.ImageIndex)
			{
			case ImageClass:
			case ImageStruct:
			case ImageInterface:
			{
				// Enumerate methods first
				MethodInfo[] methods = type.GetMethods(
					BindingFlags.Instance | BindingFlags.Public |
						BindingFlags.NonPublic |
							BindingFlags.Static | BindingFlags.DeclaredOnly);

				ConstructorInfo[] constructors = type.GetConstructors(
					BindingFlags.Instance | BindingFlags.Public |
						BindingFlags.NonPublic |
							BindingFlags.Static | BindingFlags.DeclaredOnly);

				ArrayList sortedMethods = new ArrayList();
				sortedMethods.AddRange(methods);
				sortedMethods.AddRange(constructors);
				sortedMethods.Sort(new MemberInfoComparer());

				int methodsRow = -1;

				foreach (MethodBase method in sortedMethods)
				{
					// We skip private stuff
					if (method.IsPrivate)
						continue;

					// We skip internal stuff
					if (method.IsAssembly)
						continue;

					// We process only methods here
					if (method.IsSpecialName && !method.IsConstructor)
						continue;

					int methodIcon = method.IsFamily ?  ImagePMethod : ImageMethod;

					string methodName = method.Name;
					if (method.MemberType == MemberTypes.Constructor)
						methodName = method.DeclaringType.Name;

					TreeNode childChild = new TreeNode(methodName,
						methodIcon, methodIcon);
					childChild.Tag = method;
					child.Nodes.Add(childChild);

					// Add a row to the table (if not overload)
					int col = 2;
					if (!method.IsStatic)
						col = 1;

					if (methodsRow == -1)
					{
						methodsRow = table.AddRow();
						table[0, methodsRow].ColumnSpan = 3;
						table[0, methodsRow].Text = "<b>Methods</b>";
						table.Rows[methodsRow].Header = true;
						table.Rows[methodsRow].Expanded = true;
					}
					else
					{
						bool found = false;
						for (int r = methodsRow + 1; r < table.RowCount; r++)
						{
							if (table[col, r].Text == methodName)
							{
								int overloads = (int)table[col, r].Tag + 1;
								table[col, r].ToolTip = methodName + " (+ " + overloads +
									(overloads == 1 ? " overload)" : " overloads)");
								table[col, r].Tag = overloads;

								found = true;
								break;
							}
						}

						if (found)
							continue;
					}

					int row = table.AddRow();
					if (!method.IsStatic)
						table[col, row].ColumnSpan = 2;

					table[0, row].Image = _pureImages[methodIcon];
					table[0, row].ImageAlign = ImageAlign.Center;
					table[col, row].Text = methodName;
					table[col, row].ToolTip = methodName;
					table[col, row].Tag = 0;

					if (method.IsStatic)
					{
						table[1, row].Image = _pureImages[ImageStatic];
						table[1, row].ImageAlign = ImageAlign.Center;
					}
				}

				// Enumerate properties
				PropertyInfo[] properties = type.GetProperties(
					BindingFlags.Instance | BindingFlags.Public |
						BindingFlags.NonPublic |
							BindingFlags.Static | BindingFlags.DeclaredOnly);

				ArrayList sortedProperties = new ArrayList(properties);
				sortedProperties.Sort(new MemberInfoComparer());

				int propertiesRow = -1;

				foreach (PropertyInfo property in sortedProperties)
				{
					MethodInfo getMethod = property.GetGetMethod(true);
					MethodInfo setMethod = property.GetSetMethod(true);

					if (getMethod == null && setMethod == null)
						continue;

					bool isPublic = false;
					bool isProtected = false;
					bool isStatic = false;

					if (getMethod != null)
					{
						isPublic = isPublic || getMethod.IsPublic;
						isProtected = isPublic || getMethod.IsFamily;
						isStatic = isStatic || getMethod.IsStatic;
					}

					if (setMethod != null)
					{
						isPublic = isPublic || setMethod.IsPublic;
						isProtected = isPublic || setMethod.IsFamily;
						isStatic = isStatic || setMethod.IsStatic;
					}

					if (!isPublic && !isProtected)
						continue;

					int propertyIcon = isPublic ? ImageProperty : ImagePProperty;

					TreeNode childChild = new TreeNode(property.Name,
						propertyIcon, propertyIcon);
					childChild.Tag = property;
					child.Nodes.Add(childChild);

					// Add a row to the table (if not overload)
					int col = 2;
					if (!isStatic)
						col = 1;

					if (propertiesRow == -1)
					{
						propertiesRow = table.AddRow();
						table[0, propertiesRow].ColumnSpan = 3;
						table[0, propertiesRow].Text = "<b>Properties</b>";
						table.Rows[propertiesRow].Header = true;
						table.Rows[propertiesRow].Expanded = true;
					}

					string description = "";
					foreach (Attribute attribute in property.GetCustomAttributes(false))
					{
						if (attribute is DescriptionAttribute)
						{
							description = (attribute as DescriptionAttribute).Description;
							break;
						}
					}

					int row = table.AddRow();
					if (!isStatic)
						table[col, row].ColumnSpan = 2;

					table[0, row].Image = _pureImages[propertyIcon];
					table[0, row].ImageAlign = ImageAlign.Center;
					table[col, row].Text = property.Name + " : " +
						"<color=#777777><b>" + property.PropertyType.Name + "</b></color>";
					table[col, row].ToolTip = property.Name + " : " +
						property.PropertyType.Namespace + Type.Delimiter +
							property.PropertyType.Name;
					if (description.Length > 0)
						table[col, row].ToolTip += "\n\n" + description;

					if (isStatic)
					{
						table[1, row].Image = _pureImages[ImageStatic];
						table[1, row].ImageAlign = ImageAlign.Center;
					}
				}

				// Enumerate events
				EventInfo[] events = type.GetEvents(
					BindingFlags.Instance | BindingFlags.Public |
						BindingFlags.NonPublic | BindingFlags.DeclaredOnly);

				ArrayList sortedEvents = new ArrayList(events);
				sortedEvents.Sort(new MemberInfoComparer());

				int eventsRow = -1;

				foreach (EventInfo eventInfo in sortedEvents)
				{
					MethodInfo addMethod = eventInfo.GetAddMethod(true);
					MethodInfo removeMethod = eventInfo.GetRemoveMethod(true);

					if (addMethod == null && removeMethod == null)
						continue;

					bool isPublic = false;
					bool isStatic = false;

					if (addMethod != null)
					{
						isPublic = isPublic || addMethod.IsPublic;
						isStatic = isStatic || addMethod.IsStatic;
					}

					if (removeMethod != null)
					{
						isPublic = isPublic || removeMethod.IsPublic;
						isStatic = isStatic || removeMethod.IsStatic;
					}

					if (!isPublic)
						continue;

					TreeNode childChild = new TreeNode(eventInfo.Name,
						ImageEvent, ImageEvent);
					childChild.Tag = eventInfo;
					child.Nodes.Add(childChild);

					// Add a row to the table
					int col = 2;
					if (!isStatic)
						col = 1;

					if (eventsRow == -1)
					{
						eventsRow = table.AddRow();
						table[0, eventsRow].ColumnSpan = 3;
						table[0, eventsRow].Text = "<b>Events</b>";
						table.Rows[eventsRow].Header = true;
						table.Rows[eventsRow].Expanded = true;
					}

					string description = "";
					foreach (Attribute attribute in eventInfo.GetCustomAttributes(false))
					{
						if (attribute is DescriptionAttribute)
						{
							description = (attribute as DescriptionAttribute).Description;
							break;
						}
					}

					int row = table.AddRow();

					if (!isStatic)
						table[col, row].ColumnSpan = 2;

					table[0, row].Image = _pureImages[ImageEvent];
					table[0, row].ImageAlign = ImageAlign.Center;
					table[col, row].Text = eventInfo.Name;
					table[col, row].ToolTip = eventInfo.Name;

					if (isStatic)
					{
						table[1, row].Image = _pureImages[ImageStatic];
						table[1, row].ImageAlign = ImageAlign.Center;
					}
				}

				// Enumerate fields
				if (!type.IsInterface)
				{
					FieldInfo[] fields = type.GetFields(
						BindingFlags.Instance | BindingFlags.Public |
							BindingFlags.NonPublic |
								BindingFlags.Static | BindingFlags.DeclaredOnly);

					ArrayList sortedFields = new ArrayList(fields);
					sortedFields.Sort(new MemberInfoComparer());

					int fieldsRow = -1;

					foreach (FieldInfo field in sortedFields)
					{
						if (!field.IsPublic && !field.IsFamily)
							continue;

						if (field.IsSpecialName)
							continue;

						int fieldIcon = (field.IsPublic) ? ImageField : ImagePField;

						TreeNode childChild = new TreeNode(field.Name,
							fieldIcon, fieldIcon);
						childChild.Tag = field;
						child.Nodes.Add(childChild);

						// Add a row to the table
						int col = 2;
						if (!field.IsStatic)
							col = 1;

						if (fieldsRow == -1)
						{
							fieldsRow = table.AddRow();
							table[0, fieldsRow].ColumnSpan = 3;
							table[0, fieldsRow].Text = "<b>Fields</b>";
							table.Rows[fieldsRow].Header = true;
							table.Rows[fieldsRow].Expanded = true;
						}

						int row = table.AddRow();

						if (!field.IsStatic)
							table[col, row].ColumnSpan = 2;

						table[0, row].Image = _pureImages[fieldIcon];
						table[0, row].ImageAlign = ImageAlign.Center;
						table[col, row].Text = field.Name + " : " +
							"<color=#777777><b>" + field.FieldType.Name + "</b></color>";
						table[col, row].ToolTip = field.Name + " : " +
							field.FieldType.Namespace + Type.Delimiter +
								field.FieldType.Name;
						table[col, row].Text = field.Name;
						table[col, row].ToolTip = field.Name;

						if (field.IsStatic)
						{
							table[1, row].Image = _pureImages[ImageStatic];
							table[1, row].ImageAlign = ImageAlign.Center;
						}
					}
				}
			}
				break;

			case ImageEnum:
			{
				FieldInfo[] fields = type.GetFields(
					BindingFlags.Instance | BindingFlags.Public |
						BindingFlags.NonPublic |
							BindingFlags.Static | BindingFlags.DeclaredOnly);

				ArrayList sortedFields = new ArrayList(fields);
				sortedFields.Sort(new MemberInfoComparer());

				foreach (FieldInfo field in sortedFields)
				{
					// We skip private stuff
					if (field.IsPrivate)
						continue;

					// We skip internal stuff
					if (field.IsAssembly)
						continue;

					// Skip __value field
					if (field.IsSpecialName)
						continue;

					TreeNode childChild = new TreeNode(field.Name,
						ImageEnumField, ImageEnumField);
					childChild.Tag = field;
					child.Nodes.Add(childChild);

					// Add a row to the table
					int row = table.AddRow();

					table[1, row].Image = _pureImages[ImageEnumField];
					table[1, row].ImageAlign = ImageAlign.Center;
					table[2, row].Text = field.Name;
					table[2, row].ToolTip = field.Name + " (" + (int)field.GetValue(null) + ")";
				}
			}
				break;
			}

			return child;
		}

		/// <summary>
		/// Gets the count of non-collapsed row.
		/// </summary>
		private int GetRowCount(TableNode t)
		{
			int rowCount = 0;

			bool inCollapse = false;
			for (int r = 0; r < t.RowCount; r++)
			{
				if (!inCollapse)
					rowCount++;

				if (t.Rows[r].Header)
				{
					if (!t.Rows[r].Expanded)
						inCollapse = true;
					else
						inCollapse = false;
				}
			}

			return rowCount;
		}

		private void diagram_DrawNode(object sender, DrawNodeEventArgs e)
		{
			TableNode table = e.Node as TableNode;
			if (table == null)
				return;

			IGraphics g = e.Graphics;
			Image image = _pureImages[(table.Tag as TreeNode).ImageIndex];

			// Align the destination, so that image is not blurred
			PointF dest = new PointF(e.Bounds.X + 1.5f, e.Bounds.Y + 0.5f);
			dest = diagramView.ClientToDoc(diagramView.DocToClient(dest));

			g.DrawImage(image, dest.X, dest.Y);

			// Draw scroll arrows
			Image arrow = _imageArrowUp.Image;
			Rectangle imageBounds = new Rectangle(0, 0, arrow.Width, arrow.Height);
			RectangleF imageDocBounds = diagramView.ClientToDoc(imageBounds);

			dest = new PointF(e.Bounds.Right - imageDocBounds.Width - 1.5f, e.Bounds.Y + 1);
			dest = diagramView.ClientToDoc(diagramView.DocToClient(dest));

			if (table.CurrentScrollRow > 0)
				g.DrawImage(arrow, dest.X, dest.Y);

			arrow = _imageArrowDown.Image;
			dest.Y += imageDocBounds.Height + 0.2f;
			dest = diagramView.ClientToDoc(diagramView.DocToClient(dest));

			int visibleRows = (int)((table.Bounds.Height - 7) / table.RowHeight);
			if (table.CurrentScrollRow + visibleRows < GetRowCount(table))
				g.DrawImage(arrow, dest.X, dest.Y);
		}

		private void diagram_NodeClicked(object sender, NodeEventArgs e)
		{
			TableNode table = e.Node as TableNode;
			if (table == null)
				return;

			// Check scroll arrows
			RectangleF bounds = table.Bounds;
			PointF mousePos = e.MousePosition;
			mousePos.X -= bounds.X;
			mousePos.Y -= bounds.Y;
			bounds.Offset(-bounds.X, -bounds.Y);

			Image arrow = _imageArrowUp.Image;
			Rectangle imageBounds = new Rectangle(0, 0, arrow.Width, arrow.Height);
			RectangleF imageDocBounds = diagramView.ClientToDoc(imageBounds);

			PointF dest = new PointF(bounds.Right - imageDocBounds.Width - 1.5f, bounds.Y + 1);
			dest = diagramView.ClientToDoc(diagramView.DocToClient(dest));

			if (table.CurrentScrollRow > 0)
			{
				if (new RectangleF(dest.X, dest.Y, imageDocBounds.Width, imageDocBounds.Height).Contains(mousePos))
					table.CurrentScrollRow--;
			}

			dest.Y += imageDocBounds.Height + 0.2f;
			dest = diagramView.ClientToDoc(diagramView.DocToClient(dest));

			int visibleRows = (int)((bounds.Height - 7) / table.RowHeight);
			if (table.CurrentScrollRow + visibleRows < GetRowCount(table))
			{
				if (new RectangleF(dest.X, dest.Y, imageDocBounds.Width, imageDocBounds.Height).Contains(mousePos))
					table.CurrentScrollRow++;
			}
		}


		private Comparer DefaultComparer = new Comparer(
			CultureInfo.InvariantCulture);

		private Brush ClassBrush =
			new LinearGradientBrush(Color.White, Color.LightBlue, 45);

		private Brush StructBrush =
			new LinearGradientBrush(Color.White, Color.LightPink, 45);

		private Brush InterfaceBrush =
			new LinearGradientBrush(Color.White, Color.LightGray, 45);

		private Brush EnumBrush =
			new LinearGradientBrush(Color.White, Color.FromArgb(255, 255, 160), 45);

		private Brush DelegateBrush =
			new LinearGradientBrush(Color.White, Color.PaleGreen, 45);

		private const int ImageClass = 0;
		private const int ImageDelegate = 1;
		private const int ImageEnum = 2;
		private const int ImageEnumField = 3;
		private const int ImageEvent = 4;
		private const int ImageField = 5;
		private const int ImageInterface = 6;
		private const int ImageMethod = 7;
		private const int ImagePMethod = 8;
		private const int ImagePProperty = 9;
		private const int ImageProperty = 10;
		private const int ImageStatic = 11;
		private const int ImageStruct = 12;
		private const int ImageNamespace = 13;
		private const int ImagePField = 14;

		private Image[] _pureImages;

		private Hashtable _workMap;
	}
}